import java.io.*;
import java.util.*;
import java.math.BigInteger;
import java.security.*;

public class BlockParser {
    BlockDB db;
    BlockInputStream input;
    BlockListener blklisten;
    TransactionListener txlisten;
    ArrayList<BlkInfo> blockInfos = new ArrayList<BlkInfo>();
    int txOffset;

    class BlkInfo {
	int fileNr;
	int offset;
	int depth;
    }

    class Block {
	int size;
	int version;
	byte[] parent;
	byte[] merkleHash;
	int timestamp;
	int target;
	int nonce;

	private void writeInt(MessageDigest hash, byte[] tmp, int len, long number) {
	    for(int i = 0; i < len; i++) {
		tmp[i] = (byte)(number & 0xff);
		number >>>= 8;
	    }
	    hash.update(tmp, 0, len);
	}
	
	public void serialize(MessageDigest hash) {
	    byte[] tmp = new byte[9];
	    writeInt(hash, tmp, 4, version);
	    hash.update(parent);
	    hash.update(merkleHash);
	    writeInt(hash, tmp, 4, timestamp);
	    writeInt(hash, tmp, 4, target);
	    writeInt(hash, tmp, 4, nonce);
	}
	
	public byte[] computeHash() {
	    MessageDigest sha256;
	    try {
		sha256 = MessageDigest.getInstance("SHA-256");
	    } catch (NoSuchAlgorithmException ex) {
		throw new RuntimeException(ex);
	    }

	    serialize(sha256);
	    byte[] digest = sha256.digest();
	    sha256.reset();
	    sha256.update(digest);
	    return sha256.digest();
	}
    }

    public BlockParser(TransactionListener l, String dbdir) {
	this(null, l, dbdir);
    }

    public BlockParser(BlockListener l, String dbdir) {
	this(l, null, dbdir);
    }

    public BlockParser(BlockListener bl, TransactionListener txl, String dbdir) {
	this.blklisten = bl;
	this.txlisten = txl;
	this.db = new BlockDB(dbdir);
    }

    public void readBlocks(BufferedReader reader) throws IOException {
	String s;
	int depth = 0;
	for(;;) {
	    s = reader.readLine();
	    if (s == null) {
		reader.close();
		return;
	    }
	    String[] parts = s.split(",");
	    BlkInfo info = new BlkInfo();
	    info.fileNr = Integer.parseInt(parts[0]);
	    info.offset = Integer.parseInt(parts[1]);
	    info.depth = depth++;
	    blockInfos.add(info);
	}
    }


    public void parseFromDepth(int depth) throws IOException {
	int lastFileNr = -1;
	RandomAccessFile blkfile = null;
	while (depth < blockInfos.size()) {
	    BlkInfo info = blockInfos.get(depth);
	    if (lastFileNr != info.fileNr) {
		if (blkfile != null)
		    blkfile.close();
		blkfile = db.openFile(info.fileNr);
		lastFileNr = info.fileNr;
	    }
	    blkfile.seek(info.offset);
	    input = new BlockInputStream
		(new BufferedInputStream
		 (new FileInputStream(blkfile.getFD())));
	    readBlock();
	    depth++;
	}
	if (blkfile != null)
	    blkfile.close();
    }

    public boolean readBlock() throws IOException {
	int magic = input.readInt();
	if (magic != 0xd9b4bef9 && magic != 0xc0c0c0c0)
	    throw new AssertionError("Wrong Magic");
	long size = input.readUnsignedInt();
	int blkheadstart = input.getOffset();

	Block blk = new Block();
	blk.size = (int) size;
	blk.version = (int) input.readUnsignedInt();
	blk.parent = new byte[32];
	input.readFully(blk.parent);
	blk.merkleHash = new byte[32];
	input.readFully(blk.merkleHash);
	blk.timestamp = (int) input.readUnsignedInt();
	blk.target = (int) input.readUnsignedInt();
	blk.nonce = (int) input.readUnsignedInt();
	byte[] blkhash = blk.computeHash();
	if (blklisten != null)
	    blklisten.processBlock(blk);

	if (txlisten == null)
	    return true;
	//System.out.println("New Block");
	long numTx = input.readVarInt();
	boolean coinbase = true;
	while (numTx > 0) {
	    readTransaction(coinbase);
	    numTx--;
	    coinbase = false;
	}
	return true;
    }

    public void readTransaction(boolean coinbase) throws IOException {
	Transaction tx = new Transaction();
	txOffset = input.getOffset();
	tx.isCoinBase = coinbase;
	tx.read(input);
	txlisten.processTransaction(tx);
    }
}
