import java.util.*;
import java.math.BigInteger;
import java.io.*;
import java.security.*;

public class LongestChain {
    HashMap<BigInteger,Block> blocks = new LinkedHashMap<BigInteger,Block>();
    Block bestBlock;
    BlockDB db;
    int fileNr;
    int maxFileNr;

    class Block {
	Block parentBlock;
	int fileNr;
	int offset;
	int size;
	int version;
	byte[] parent;
	byte[] merkleHash;
	int timestamp;
	int target;
	int nonce;
	int depth = -1;
	byte[] hash;

	private void writeInt(MessageDigest hash, byte[] tmp, int len, long number) {
	    for(int i = 0; i < len; i++) {
		tmp[i] = (byte)(number & 0xff);
		number >>>= 8;
	    }
	    hash.update(tmp, 0, len);
	}
	
	public void serialize(MessageDigest hash) {
	    byte[] tmp = new byte[4];
	    writeInt(hash, tmp, 4, version);
	    hash.update(parent);
	    hash.update(merkleHash);
	    writeInt(hash, tmp, 4, timestamp);
	    writeInt(hash, tmp, 4, target);
	    writeInt(hash, tmp, 4, nonce);
	}
	
	public byte[] computeHash() {
	    if (hash != null)
		return hash;

	    MessageDigest sha256;
	    try {
		sha256 = MessageDigest.getInstance("SHA-256");
	    } catch (NoSuchAlgorithmException ex) {
		throw new RuntimeException(ex);
	    }

	    serialize(sha256);
	    byte[] digest = sha256.digest();
	    sha256.reset();
	    sha256.update(digest);
	    hash = sha256.digest();
	    return hash;
	}
    }

    public LongestChain(String dbdir) {
	this.db = new BlockDB(dbdir);
    }

    byte[] blockheader = new byte[4+4+80];
    public boolean readBlockHeader(RandomAccessFile file, Block blk) throws IOException {
	int offset = (int) file.getFilePointer();
	try {
	    int skip = 0;
	    read:
	    for (;;) {
		file.readFully(blockheader, skip, blockheader.length - skip);
		if (blockheader[0] != 0)
		    break;
		// skip zeros
		for (int i = 0; i < blockheader.length; i++) {
		    if (blockheader[i] != 0) {
			skip = blockheader.length - i;
			System.arraycopy(blockheader, i, blockheader, 0, skip);
			offset += i;
			continue read;
		    }
		}
		offset += blockheader.length;
		skip = 0;
	    }
	} catch (EOFException ex) {
	    return false;
	}
	BlockInputStream input = new BlockInputStream
	    (new ByteArrayInputStream(blockheader));
	int magic = input.readInt();
	if (magic != 0xd9b4bef9 && magic != 0xc0c0c0c0)
	    throw new AssertionError("Wrong Magic");
	blk.offset = offset;
	readBlockHeader(input, blk);
	return true;
    }


    public void readBlockHeader(BlockInputStream input, Block blk) throws IOException {
	blk.size = (int) input.readUnsignedInt();
	blk.version = (int) input.readUnsignedInt();
	blk.parent = new byte[32];
	input.readFully(blk.parent);
	blk.merkleHash = new byte[32];
	input.readFully(blk.merkleHash);
	blk.timestamp = (int) input.readUnsignedInt();
	blk.target = (int) input.readUnsignedInt();
	blk.nonce = (int) input.readUnsignedInt();
	byte[] blkhash = blk.computeHash();
	Block prevblk = blocks.get(new BigInteger(1, blkhash));
	if (prevblk == null)
	    blocks.put(new BigInteger(1, blkhash), blk);
    }

    public boolean readBlock(RandomAccessFile file) throws IOException {
	int offset = (int) file.getFilePointer();
	
	Block blk = new Block();
	blk.fileNr = fileNr;
	if (!readBlockHeader(file, blk))
	    return false;
	file.skipBytes(blk.size-80);	// skip transactions

	//System.err.println("Block: "+Util.hexBytes(Util.reverse(blk.computeHash())) + " parent: " + Util.hexBytes(Util.reverse(blk.parent)));
	return true;
    }
    
    public void computeDepths() {
	ArrayDeque<Block> todo = new ArrayDeque<Block>();
	todo.addAll(blocks.values());
	while (!todo.isEmpty()) {
	    Block blk = todo.removeFirst();
	    /* compute depth */
	    if (blk.depth != -1) {
		// nothing
	    } else if (Arrays.equals(blk.parent, new byte[32])) {
		blk.depth = 0;
	    } else {
		Block parent = blocks.get(new BigInteger(1,blk.parent));
		//System.err.println("Blk: "+Util.hexBytes(Util.reverse(blk.computeHash())) + " P: "+Util.hexBytes(Util.reverse(blk.parent)));
		if (parent == null)  {
		    System.err.println("Parent block not found: "+Util.hexBytes(Util.reverse(blk.parent)));
		    blk.depth = -100000000;
		} else if (parent.depth == -1) {
		    todo.addFirst(blk);
		    todo.addFirst(parent);
		} else {
		    blk.depth = parent.depth+1;
		    blk.parentBlock = parent;
		}
	    }
	    if (bestBlock == null
		|| blk.depth > bestBlock.depth)
		bestBlock = blk;
	}
    }
    
    public void dumpBlockChain() {
	computeDepths();
	ArrayDeque<Block> chain = new ArrayDeque<Block>();
	Block blk = bestBlock;
	while (blk != null) {
	    chain.addFirst(blk);
	    blk = blk.parentBlock;
	}
	for (Block block : chain) {
	    System.out.println(block.fileNr + "," + block.offset);
	}
    }


    public void readBlocks(BufferedReader reader) throws IOException {
	String s;
	int depth = 0;
	Block lastBlock = null;
	for(;;) {
	    s = reader.readLine();
	    if (s == null) {
		reader.close();
		break;
	    }
	    String[] parts = s.split(",");
	    int fileNr = Integer.parseInt(parts[0]);
	    int offset  = Integer.parseInt(parts[1]);
	    //System.err.println(s);
	    Block blk = new Block();
	    blk.fileNr = fileNr;
	    blk.offset = offset;
	    blk.depth = depth++;
	    blk.parentBlock = lastBlock;
	    lastBlock = blk;
	}
	bestBlock = lastBlock;
	RandomAccessFile blkfile = db.openFile(0);
	int lastFileNr = 0;
	for (int i = 0; i < 20; i++) {
	    Block blk = lastBlock;
	    if (blk == null)
		break;
	    if (blk.fileNr != lastFileNr) {
		blkfile.close();
		blkfile = db.openFile(blk.fileNr);
		lastFileNr = blk.fileNr;
	    }
	    blkfile.seek(blk.offset);
	    BlockInputStream input = 
		new BlockInputStream
		(new BufferedInputStream
		 (new FileInputStream(blkfile.getFD())));
	    int magic = input.readInt();
	    if (magic != 0xd9b4bef9 && magic != 0xc0c0c0c0) {
		throw new AssertionError("Wrong Magic: "
					 + blk.fileNr + "," + blk.offset);
	    }
	    readBlockHeader(input, blk);
	    lastBlock = lastBlock.parentBlock;
	    if (blk.fileNr > maxFileNr)
		maxFileNr = blk.fileNr;
	}
	blkfile.close();
    }

    public static void main(String[] param) throws IOException {
	LongestChain lc = new LongestChain(param[1]);
	lc.readBlocks(new BufferedReader(new FileReader(param[0])));
	lc.expand();
    }

    public void expand() throws IOException {
	fileNr = Math.max(0, this.maxFileNr - 1);
	while (true) {
	    RandomAccessFile file;
	    try {
		file = db.openFile(fileNr);
	    } catch (FileNotFoundException ex) {
		break;
	    }
	    System.err.println("Block " + fileNr);
	    while (readBlock(file)) {
		// new block
	    }
	    file.close();
	    fileNr++;
	}
	dumpBlockChain();
    }
}
