import java.io.*;

public class SegWitEstimator implements TransactionListener, BlockListener {
    TransactionReader txreader;
    public long origsize;
    public long p2shsavings;
    public long segwitsavings;
    public int transactions;
    public int p2hashInputs;
    public int p2hashOutputs;
    public int p2shInputs;
    public int p2shOutputs;
    public int blockNr;

    public SegWitEstimator(TransactionReader txr) {
	txreader = txr;
    }

    public boolean isDERSignature(byte[] der) {
	return Util.decodeDERSignature(der) != null;
    }

    public boolean isPublicKey(byte[] key) {
	if (key.length == 65 && key[0] == 4)
	    return true;
	if (key.length == 33 && (key[0] & ~1) == 2)
	    return true;
	return false;
    }

    public boolean isKnownScript(byte[] script) {
	Script oscript = new Script(script);
	return oscript.isCheckSig() || oscript.isCheckMultiSig()
	    || oscript.isPayToHash();
    }
    
    public void processTransaction(Transaction tx) {
	int size, segwitsavein, p2shsavein, segwitsaveout, p2shsaveout;
	size = tx.computeSize();
	origsize += size;
	transactions++;
	// account for segwitheader + numwitness + witnessheaders
	// for simplicity assume that every size field fits in one byte. 
	segwitsavein = -3 - tx.inputs.length;
	p2shsavein = -3 - tx.inputs.length;
	if (!tx.isCoinBase) {
	    for (int i = 0; i < tx.inputs.length; i++) {
		byte[] script = tx.inputs[i].scriptParam;
		byte[][] inputs = Util.decodeScriptInput(script);
		
		if (inputs.length == 2
		    && isDERSignature(inputs[0])
		    && isPublicKey(inputs[1])) {
		    // This looks like a standard P2PKH input
		    // segwit moves input script to witness data
		    segwitsavein += 4 * script.length
			- (script.length);
		    // p2sh needs an additional witness script in
		    // input that is not discounted.
		    p2shsavein += 4 * script.length
			- (script.length)
			- 4 * 22; // witness address encoded in scriptParams
		    p2hashInputs++;
		} else if (isKnownScript(inputs[inputs.length-1])) {
		    // segwit moves input script to witness data
		    segwitsavein += 4 * script.length
			- (script.length);
		    // p2sh needs an additional witness script in
		    // input that is not discounted.
		    p2shsavein += 4 * script.length
			- (script.length)
			- 4 * 34; // witness p2sh encoded in scriptParams
		    p2shInputs++;
		} else {
		    try {
			Transaction txold = txreader.getTransaction(tx.inputs[i].hash);
			byte[] scripto = txold.outputs[tx.inputs[i].outIdx].script;
			Script oscript = new Script(scripto);
			if (oscript.isCheckSig()) {
			    segwitsavein += 4 * (script.length)
				- script.length - (scripto.length - 3);
			    p2shsavein += 4 * (script.length)
				- script.length - (scripto.length - 3)
				- 4 * 22;
			    p2hashInputs++;
			} else if (oscript.isCheckMultiSig()) {
			    segwitsavein += 4 * (script.length)
				- script.length - scripto.length;
			    p2shsavein += 4 * (script.length)
				- script.length - scripto.length
				- 4 * 34;
			    p2shInputs++;
			} else {
			    System.err.println("Unknown tx input: "+Util.hexBytes(Util.reverse(tx.computeHash()))+":"+i);
			    // ignore savings
			}
		    } catch (IOException ex) {
			System.err.println("Missing TX: " 
					   + Util.hexBytes(Util.reverse(tx.inputs[i].hash)));
		    }
		}
	    }
	    segwitsavings += segwitsavein;
	    p2shsavings += p2shsavein;
	}


	{
	    segwitsaveout = 0;
	    p2shsaveout = 0;
	    for (int i = 0; i < tx.outputs.length; i++) {
		byte[] script = tx.outputs[i].script;
		Script oscript = new Script(script);
		if (oscript.isPayToHash()) {
		    segwitsaveout += 4 * (script.length - 22);
		    p2shsaveout += 4 * (script.length - 23);
		    p2hashOutputs++;
		} else if (oscript.isPayToScriptHash()) {
		    segwitsaveout += 4 * (script.length - 34);
		    p2shsaveout += 4 * (script.length - 23);
		    p2shOutputs++;
		} else if (oscript.isCheckSig()) {
		    segwitsaveout += 4 * (script.length - 22);
		    p2shsaveout += 4 * (script.length - 23);
		    p2hashOutputs++;
		} else if (oscript.isCheckMultiSig()) {
		    segwitsaveout += 4 * (script.length - 34);
		    p2shsaveout += 4 * (script.length - 23);
		    p2shOutputs++;
		} else {
		    if (!oscript.isOpReturn()) {
			System.err.println("Unknown output: "+oscript);
		    }
		}
	    }
	    segwitsavings += segwitsaveout;
	    p2shsavings += p2shsaveout;
	}

	int segwitsize = size - (segwitsavein + segwitsaveout) / 4;
	int p2shsize = size - (p2shsavein + p2shsaveout) / 4;
	//System.out.println("tx: "+size+"/"+p2shsize+"/"+segwitsize);
    }

    public void summary() {
	long segwitsize = origsize + (3 - segwitsavings) / 4; // round up
	long p2shsize = origsize + (3 - p2shsavings) / 4;     // round up
	System.out.println(""+blockNr+";"+transactions+";"+
			   origsize + ";" + 
			   segwitsize + ";" + 
			   p2shsize + ";" +
			   (double)origsize / segwitsize + ";" + 
			   (double)origsize / p2shsize);
			   
	// System.out.println("TOTAL " + transactions + " Transactions");
	// System.out.println("  p2pkh " + p2hashInputs + " / " + p2hashOutputs);
	// System.out.println("  p2sh  " + p2shInputs + " / " + p2shOutputs);
	// System.out.println("  "+origsize+"/"+p2shsize+"/"+segwitsize);
	// System.out.println("  Savings: "+ (100.0 * (origsize - p2shsize)/origsize) +" % / " +
	// 		   (100.0 * (origsize - segwitsize)/origsize) +" %");

	transactions = 0;
	origsize = 80;
	p2shsavings = 0;
	segwitsavings = 0;
    }

    public void processBlock(BlockParser.Block blk) {
	if (transactions > 0)
	    summary();
	blockNr++;
    }
    
    public static void main(String[] param) throws IOException {
	System.out.println("\"Block Nr\";\"transactions\";\"orig size\";\"segwit size\";\"p2sh segwit size\";\"segwit factor\";\"p2sh segwit factor\"");
	SegWitEstimator me = new SegWitEstimator(new TransactionReader(param[0]));
	BlockParser bp = new BlockParser(me, me, param[2]);
	bp.readBlocks(new BufferedReader(new FileReader(param[1])));	
	int blockNr = Integer.parseInt(param[3]);
	me.blockNr = blockNr - 1;
	bp.parseFromDepth(blockNr);
	me.summary();
    }
}
