import java.security.*;
import java.io.*;

public class Transaction {
    boolean isCoinBase;
    int version;
    Input[] inputs;
    Output[] outputs;
    int lockTime;

    class Input {
	byte[] hash;
	int    outIdx;
	byte[] scriptParam;
	int    sequenceNo;

	public Input(Input copy) {
	    hash = copy.hash;
	    outIdx = copy.outIdx;
	    scriptParam = copy.scriptParam;
	    sequenceNo = copy.sequenceNo;
	}
	public Input() {}

	public String toString() {
	    StringBuilder sb = new StringBuilder();
	    sb.append("{\"tx\":\"").append(Util.hexBytes(hash));
	    sb.append("\",\"out\":").append(outIdx);
	    sb.append(",\"script\":\"").append(Util.hexBytes(scriptParam));
	    sb.append("\",\"sequence\":").append(sequenceNo);
	    sb.append("}");
	    return sb.toString();
	}
    }

    class Output {
	long   amount;
	byte[] script;

	public Output(Output copy) {
	    amount = copy.amount;
	    script = copy.script;
	}
	public Output() {}

	public String toString() {
	    StringBuilder sb = new StringBuilder();
	    sb.append("{\"script\":\"").append(Util.hexBytes(script));
	    sb.append("\",\"amount\":").append(amount);
	    sb.append("}");
	    return sb.toString();
	}
    }

    public Transaction() {}
    public Transaction(Transaction copy) {
	version = copy.version;
	inputs = new Input[copy.inputs.length];
	for (int i = 0; i < inputs.length; i++)
	    inputs[i] = new Input(copy.inputs[i]);
	outputs = new Output[copy.outputs.length];
	for (int i = 0; i < outputs.length; i++)
	    outputs[i] = new Output(copy.outputs[i]);
	lockTime = copy.lockTime;
    }

    public void read(BlockInputStream input) throws IOException {
    	version = input.readInt();
	int incount = (int)input.readVarInt();
	inputs = new Input[(int)incount];
	for (int i = 0; i < incount; i++) {
	    inputs[i] = readInput(input);
	}
	int outcount = (int) input.readVarInt();
	outputs = new Output[outcount];
	for (int i = 0; i < outcount; i++) {
	    outputs[i] = readOutput(input);
	}
	lockTime = input.readInt();
	if (version != 1)
	    System.err.println("Warning: Wrong version: " + version + " in "
			       + Util.hexBytes(Util.reverse(computeHash())));
    }

    private Input readInput(BlockInputStream input) throws IOException {
	Input in = new Input();
    	in.hash = new byte[32];
	input.readFully(in.hash);
	in.outIdx = (int) input.readUnsignedInt();
	int skriptLen = (int) input.readVarInt();
	in.scriptParam = new byte[skriptLen];
	input.readFully(in.scriptParam);
	in.sequenceNo = (int) input.readUnsignedInt();
	return in;
    }

    private Output readOutput(BlockInputStream input) throws IOException {
	Output out = new Output();
    	out.amount = input.readUnsignedLong();
	int skriptLen = (int) input.readVarInt();
	out.script = new byte[skriptLen];
	input.readFully(out.script);
	return out;
    }

    private void writeInt(MessageDigest hash, byte[] tmp, int len, long number) {
	for(int i = 0; i < len; i++) {
	    tmp[i] = (byte)(number & 0xff);
	    number >>>= 8;
	}
	hash.update(tmp, 0, len);
    }
    private void writeVarInt(MessageDigest hash, byte[] tmp, long number) {
	int len;
	if (number >=0 && number < 0xfd) {
	    tmp[0] = (byte)number;
	    len = 1;
	} else if (number <= 0xffff) {
	    tmp[0] = (byte)0xfd;
	    len = 3;
	} else if (number <= 0xffffffff) {
	    tmp[0] = (byte)0xfe;
	    len = 5;
	} else {
	    tmp[0] = (byte)0xff;
	    len = 9;
	}
	for(int i = 1; i < len; i++) {
	    tmp[i] = (byte)(number & 0xff);
	    number >>>= 8;
	}
	hash.update(tmp, 0, len);
    }

    public void serialize(MessageDigest hash) {
	byte[] tmp = new byte[9];
	writeInt(hash, tmp, 4, version);
	writeVarInt(hash, tmp, inputs.length);
	for (int i = 0; i < inputs.length; i++) {
	    hash.update(inputs[i].hash);
	    writeInt(hash, tmp, 4, inputs[i].outIdx);
	    writeVarInt(hash, tmp, inputs[i].scriptParam.length);
	    hash.update(inputs[i].scriptParam);
	    writeInt(hash, tmp, 4, inputs[i].sequenceNo);
	}
	writeVarInt(hash, tmp, outputs.length);
	for (int i = 0; i < outputs.length; i++) {
	    writeInt(hash, tmp, 8, outputs[i].amount);
	    writeVarInt(hash, tmp, outputs[i].script.length);
	    hash.update(outputs[i].script);
	}
	writeInt(hash, tmp, 4, lockTime);
    }

    public int varIntSize(int number) {
	if (number >=0 && number < 0xfd) {
	    return 1;
	} else if (number <= 0xffff) {
	    return 3;
	} else if (number <= 0xffffffff) {
	    return 5;
	} else {
	    return 9;
	}
    }

    public int computeSize() {
	int size = 4 + 4; // version + lockTime;
	size += varIntSize(inputs.length);
	for (int i = 0; i < inputs.length; i++) {
	    size += 32 + 4 + 4 + 
		varIntSize(inputs[i].scriptParam.length) +
		inputs[i].scriptParam.length;
	}
	size += varIntSize(outputs.length);
	for (int i = 0; i < outputs.length; i++) {
	    size += 8 +
		varIntSize(outputs[i].script.length) +
		outputs[i].script.length;
	}
	return size;
    }

    public byte[] computeHash() {
	MessageDigest sha256;
	try {
	    sha256 = MessageDigest.getInstance("SHA-256");
	} catch (NoSuchAlgorithmException ex) {
	    throw new RuntimeException(ex);
	}
	/*
	sha256 = new MessageDigest("foo") {
		protected void engineUpdate(byte data) {
			System.err.println(Integer.toString(data, 16));
		}
		protected void engineUpdate(byte[] data, int offset, int len) {
		    for (int i = 0; i< len; i++)
			System.err.print(Integer.toString(0x200 + data[offset+i], 16).substring(1)+" ");
		}
		protected void engineReset() {}
		protected byte[] engineDigest(){
		    return new byte[32];
		}
	    };
	*/
	serialize(sha256);
	byte[] digest = sha256.digest();
	sha256.reset();
	sha256.update(digest);
	return sha256.digest();
    }

    public String toString() {
	StringBuilder sb = new StringBuilder();
	sb.append("{\"version\":").append(version);
	sb.append(",\"inputs\":[");
	String comma = "";
	for (int i = 0; i < inputs.length; i++) {
	    sb.append(comma).append(inputs[i].toString());
	    comma = ",";
	}
	sb.append("],\"outputs\":[");
	comma = "";
	for (int i = 0; i < outputs.length; i++) {
	    sb.append(comma).append(outputs[i].toString());
	    comma = ",";
	}
	sb.append("],\"locktime\":").append(lockTime);
	sb.append("}");
	return sb.toString();
    }
}
